/*****************************************************************************
 * Copyright (c) 2007 Rowley Associates Limited.                             *
 *                                                                           *
 * This file may be distributed under the terms of the License Agreement     *
 * provided with this software.                                              *
 *                                                                           *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE   *
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. *
 *****************************************************************************/

/*****************************************************************************
 *                           Preprocessor Definitions
 *                           ------------------------
 *
 * STARTUP_FROM_RESET
 *
 *   If defined, the program will startup from power-on/reset. If not defined
 *   the program will just loop endlessly from power-on/reset.
 *
 *   This definition is not defined by default on this target because the    
 *   debugger is unable to reset this target and maintain control of it over the
 *   JTAG interface. The advantage of doing this is that it allows the debugger
 *   to reset the CPU and run programs from a known reset CPU state on each run.
 *   It also acts as a safety net if you accidently download a program in FLASH
 *   that crashes and prevents the debugger from taking control over JTAG
 *   rendering the target unusable over JTAG. The obvious disadvantage of doing
 *   this is that your application will not startup without the debugger.
 *
 *   We advise that on this target you keep STARTUP_FROM_RESET undefined whilst
 *   you are developing and only define STARTUP_FROM_RESET when development is
 *   complete.
 *
 * EMI_BCONn_VAL
 *
 *   If defined enable P2.0..P2.7 pins for EMI AF and 
 *   set the EMI_BCONn_VAL external memory bank control register value.
 *
 *****************************************************************************/

 #include <targets/STR71x.h>

  .section .vectors, "ax"
  .code 32
  .align 0
  .global _vectors
  .global reset_handler
  
/*****************************************************************************
 * Exception Vectors                                                         *
 *****************************************************************************/
_vectors:
#ifdef STARTUP_FROM_RESET
  ldr pc, [pc, #reset_handler_address - . - 8]  /* reset */
#else
  b .                                           /* reset - infinite loop */
#endif
  ldr pc, [pc, #undef_handler_address - . - 8]  /* undefined instruction */
  ldr pc, [pc, #swi_handler_address - . - 8]    /* swi handler */
  ldr pc, [pc, #pabort_handler_address - . - 8] /* abort prefetch */
  ldr pc, [pc, #dabort_handler_address - . - 8] /* abort data */
  nop
  ldr pc, [pc, #irq_handler_address - . - 8]    /* irq */
  ldr pc, [pc, #fiq_handler_address - . - 8]    /* fiq */

reset_handler_address:
  .word reset_handler
undef_handler_address:
  .word undef_handler
swi_handler_address:
  .word swi_handler
pabort_handler_address:
  .word pabort_handler
dabort_handler_address:
  .word dabort_handler
irq_handler_address:
  .word tn_cpu_irq_isr
fiq_handler_address:
  .word tn_cpu_fiq_isr

  .section .init, "ax"
  .code 32
  .align 0

#ifndef PROCESSOR_CLOCK_FREQUENCY
#define PROCESSOR_CLOCK_FREQUENCY 48000000
#endif

/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 *                                                                            *
 ******************************************************************************/
reset_handler:
#ifdef __FLASH_BUILD
  // Copy exception vectors into RAM 
  mov r8, #0x20000000
  ldr r9, =_vectors
  ldmia r9!, {r0-r7}
  stmia r8!, {r0-r7}
  ldmia r9!, {r0-r6}
  stmia r8!, {r0-r6}     
#endif

#if defined(EMI_BCON0_VAL) || defined(EMI_BCON1_VAL) || defined(EMI_BCON2_VAL) || defined(EMI_BCON3_VAL)
  // Set P2.0..P2.7 to EMI function
  ldr r0, =IOPORT2_BASE
  ldr r1, =0x000000FF
  str r1, [r0, #IOPORT2_PC2_OFFSET]

  ldr r0, =EMI_BASE
#endif
#if defined(EMI_BCON0_VAL)
  ldr r1, =EMI_BCON0_VAL
  strh r1, [r0, #EMI_BCON0_OFFSET]
#endif
#if defined(EMI_BCON1_VAL)
  ldr r1, =EMI_BCON1_VAL
  strh r1, [r0, #EMI_BCON1_OFFSET]
#endif
#if defined(EMI_BCON2_VAL)
  ldr r1, =EMI_BCON2_VAL
  strh r1, [r0, #EMI_BCON2_OFFSET]
#endif
#if defined(EMI_BCON3_VAL)
  ldr r1, =EMI_BCON3_VAL
  strh r1, [r0, #EMI_BCON3_OFFSET]
#endif

  // Map internal RAM to 0x00000000
  ldr r0, =PRCCU_BASE
  ldr r1, [r0, #PCU_BOOTCR_OFFSET]
  bic r1, r1, #0x01
  orr r1, r1, #0x02
  str r1, [r0, #PCU_BOOTCR_OFFSET]

  // Note that PLL works from CLK2 which is CLK div 2 by default
#if OSCILLATOR_CLOCK_FREQUENCY==16000000  
#if PROCESSOR_CLOCK_FREQUENCY==32000000
  ldr r1, =0x73            // set the PLL to x16 and div 4 == 32Mhz
#elif PROCESSOR_CLOCK_FREQUENCY==48000000
  ldr r1, =0x51            // set the PLL to x12 and div 2 == 48Mhz
#endif
#elif OSCILLATOR_CLOCK_FREQUENCY==4000000
#if PROCESSOR_CLOCK_FREQUENCY==32000000
  ldr r1, =0x30            // set the PLL to x16 and div 1 == 32Mhz 
#elif PROCESSOR_CLOCK_FREQUENCY==48000000
  ldr r1, =0x20              // set the PLL to x24 and div 1 == 48Mhz
#endif
#else
#error OSCILLATOR_CLOCK_FREQUENCY not defined
#endif

  str r1, [r0, #RCCU_PLL1CR_OFFSET]
  ldr r1, =0x8008 /* Enable PLL */
  str r1, [r0, #RCCU_CFR_OFFSET]
1:
  ldr r1, [r0, #RCCU_CFR_OFFSET]
  tst r1, #0x0002
  beq 1b /* Wait for PLL to lock */             

  orr r1, r1, #1 /* Select PLL as system clock */ 
  str r1, [r0, #RCCU_CFR_OFFSET] 
  
  ldr r1, =0x0000 /* CPU clock divider = 1 */
  str r1, [r0, #PCU_MDIVR_OFFSET]

#if PROCESSOR_CLOCK_FREQUENCY==32000000
  ldr r1, =0x0000 /* APB clock dividers = 1  */
#else
  ldr r1, =0x0101 /* APB clock dividers = 2  */
#endif
  str r1, [r0, #PCU_PDIVR_OFFSET]

  /* Jump to the default C runtime startup code. */
  b _start

/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 * These are declared weak symbols so they can be redefined in user code.     * 
 *                                                                            *
 ******************************************************************************/

undef_handler:
  b .  /* Endless loop */
  
swi_handler:
  b .  /* Endless loop */
  
pabort_handler:
  b .  /* Endless loop */
  
dabort_handler:
  b .  /* Endless loop */
  
irq_handler:
  b .  /* Endless loop */
  
fiq_handler:
  b .  /* Endless loop */

  .weak undef_handler, swi_handler, pabort_handler, dabort_handler, irq_handler, fiq_handler
