/****************************************************************************
*  This file is part of the str71x port for the TNKernel RTOS
*
*  Copyright (c) 2009 by Michael Fischer. All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
****************************************************************************
*  History:
*
*  20.09.2009  mifi   First Version
****************************************************************************/
#if defined(TN_ARCH_ARM) && defined(TN_PORT_STR71X)

/*=========================================================================*/
/*  Includes                                                               */
/*=========================================================================*/
#include <targets/str71x.h>
#include "tn.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
typedef void (*IRQ_FUNC)(void);

/*=========================================================================*/
/*  DEFINE: Definition of all local Data                                   */
/*=========================================================================*/
static IRQ_FUNC IRQTable[TN_MAX_IRQ_ID];

static unsigned long *SourceInterruptTable = (unsigned long*)(EIC_BASE + EIC_SIR0_OFFSET);

/*=========================================================================*/
/*  DEFINE: Definition of all local Procedures                             */
/*=========================================================================*/

/***************************************************************************/
/*  DefaultIRQHandler                                                      */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: never                                                          */
/***************************************************************************/
static void DefaultIRQHandler (void)
{
   volatile int i = 0;
   
   while(1)
   {
      i++;
   }
} /* DefaultIRQHandler */

/*=========================================================================*/
/*  DEFINE: All code exported                                              */
/*=========================================================================*/

/***************************************************************************/
/*  tn_cpu_irq_handler                                                     */
/*                                                                         */
/*  This function will execute the user interrupt function.                */
/*                                                                         */                                                                       
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
void tn_cpu_irq_handler (void)
{
   int      idx;
   IRQ_FUNC IrqFunc;
      
   /*
    * The EIC_IVR is the EIC register that the sofware has to read after 
    * detecting the nIRQ signal assertion. The EIC_IVR read operation informs 
    * the EIC that the interrupt service routine (ISR) corresponding to the 
    * pending request has been initiated.
    *
    * This means that:
    *  - the IRQ signal can be de-asserted
    *  - the EIC_CIPR and EIC_CICR can be updated
    *  - no interrupt requests, whose priority is lower or 
    *    equal than the current one can be processed.
    */
    
   /* Get the index in the IRQTable */ 
   idx = (EIC_IVR & 0x1F);
   
   /* Get the IRQ function */
   IrqFunc = IRQTable[idx];   
   if (IrqFunc != (IRQ_FUNC)0)
   {
      IrqFunc();
   }
   
   /*
    * Before exiting an ISR, the software must have cleared the EIC_IPR0 
    * bit related to the executed routine. This bit clear operation will be 
    * interpreted by the EIC as End of Interrupt (EOI) sequence and will 
    * allow the interrupt stack pop and processing of new interrupts.
    */
   EIC_IPR0 |= (1L << idx);    
   
} /* tn_cpu_irq_handler */

/***************************************************************************/
/*  tn_irq_init                                                            */
/*                                                                         */
/*  Here the irq system of the cpu will be init.                           */
/*  E.g. a spurious interrupt will be set.                                 */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
void tn_irq_init (void)
{
   int i;
   
   /* Disable both IRQ & FIR */
   tn_arm_disable_interrupts();

   /* Disable all interrupts first */
   EIC_IER0 = 0;
   
   /* Disable FIQ and IRQ */   
   EIC_ICR = 0;
   
   /* Clear all pending interrupts */
   EIC_IPR0 = 0xFFFFFFFF;
   
   /* Set IRQTable to 0 */
   for(i=0; i<TN_MAX_IRQ_ID; i++)
   {
      IRQTable[i] = DefaultIRQHandler;
   }
   
} /* tn_irq_init */

/***************************************************************************/
/*  tn_irq_install                                                         */
/*                                                                         */
/*  In    : id, prio, interrupt function                                   */
/*  Out   : none                                                           */
/*  Return: TERR_NO_ERR / TERR_WRONG_PARAM                                 */
/***************************************************************************/
int tn_irq_install (TN_IRQ_ID   id,
                    TN_IRQ_PRIO prio,
                    void (*irq_func)(void))
{
   int rc = TERR_NO_ERR;
   
   if( (id >= TN_MAX_IRQ_ID) || (prio >= TN_MAX_IRQ_PRIO) )
   {
      rc = TERR_WRONG_PARAM;
   }
   else
   {
      /* Disable interrupt first */
      EIC_IER0 &= ~(1L << id);
   
      IRQTable[id] = irq_func;
      SourceInterruptTable[id] = ((unsigned long)(id << 16)) | prio;
   }
   
   return(rc);
} /* tn_irq_install */

#endif /* TN_ARCH_ARM && TN_PORT_STR71X */
/*** EOF ***/
